from .constant import api
import time
from requests import Session
from .signalr import Connection
from . import FCTradingClient


class FCTradingStream(object):

    def __init__(self, fctrading_client: FCTradingClient, stream_url: str, on_message, on_error, last_notify_id: str = "-1", on_close=None, on_open=None):
        self.client = fctrading_client
        self._stream_url = stream_url
        self._fc_handlers = []
        self._fc_handlers.append(on_message)
        self._fc_error_handlers = []
        self._fc_error_handlers.append(on_error)
        self._on_close = on_close
        self._on_open = on_open
        self.lastId = last_notify_id

    def fc_on_message(self, fc_message):
        for fc_handler in self._fc_handlers:
            fc_handler(fc_message)

    def fc_on_error(self, fc_error):

        for fc_error_handler in self._fc_error_handlers:
            fc_error_handler(fc_error)

    def on_close(self):
        if self._on_close and callable(self._on_close):
          self._on_close()

    def on_open(self):
      if self._on_open and callable(self._on_open):
          self._on_open()

    def start(self):

        headers = {}
        headers['Authorization'] = "Bearer " + self.client.get_access_token()
        headers["NotifyID"] = self.lastId

        self.connection = Connection(self._stream_url + api.SIGNALR, headers, on_close=self.on_close, on_open=self.on_open)

        self.hub_proxy = self.connection.register_hub(api.SIGNALR_HUB_FC)

        self.hub_proxy.client.on(api.SIGNALR_METHOD, self.fc_on_message)

        self.hub_proxy.client.on(api.SIGNALR_METHOD_ERROR, self.fc_on_error)

        self.connection.error += self.fc_on_error

        self.connection.start()



from ssi_fctrading import FCTradingClient
from ssi_fctrading.models import fcmodel_requests
from . import fc_config
from typing import Union

from fastapi import FastAPI, Depends
app = FastAPI()

import random
client = FCTradingClient(fc_config.Url, fc_config.ConsumerID,
	fc_config.ConsumerSecret, fc_config.PrivateKey, fc_config.TwoFAType)
print('Read token: ' + client.get_access_token())
@app.get("/getOtp")
async def fc_get_otp():
	"""Get OPT if you use SMS OTP or Email OTP

	Returns:
		string: response json string
	"""
	fc_req = fcmodel_requests.GetOTP(fc_config.ConsumerID, fc_config.ConsumerSecret)
	return client.get_otp(fc_req)

@app.get("/verifyCode")
async def fc_get_otp(code: str):
	"""Verify OTP or PIN (with TwoFAType in your config), if you use SMS OTP or Email OTP please get call getOtp to receive OTP before verify.
	 This function auto save OTP and accesstoken for New/Modify/Cancel order

	Returns:
		string: response json string
	"""
	return client.verifyCode(code)

@app.get("/newOrder")
async def fc_new_order(instrumentID: str, market: str, buySell: str, orderType: str
    , price: float, quantity: int, account: str, stopOrder: bool = False, stopPrice: float = 0, 
      stopType: str = '', stopStep: float = 0, lossStep: float = 0, profitStep: float = 0, deviceId: str = FCTradingClient.get_deviceid(), userAgent: str = FCTradingClient.get_user_agent()):
	
	"""Place new order

	Args:
	```	
	instrumentID (str): Mã chứng khoán
	market (str): Thị trường ('VN' hoặc 'VNFE')
	buySell (str): 'B' or 'S'
	orderType (str): Loại lệnh
	price (float): Giá. Với các lệnh điều kiện price=0
	quantity (int): Khối lượng
	account (str): Tài khoản
	stopOrder (bool, optional): Lệnh điều kiện (chỉ áp dụng với phái sinh). Defaults to False.
	stopPrice (float, optional): Giá trigger của lệnh điều kiện. Defaults to 0.
	stopType (str, optional): Loại lệnh điều kiện. Defaults to ''.
	stopStep (float, optional): . Defaults to 0.
	lossStep (float, optional): . Defaults to 0.
	profitStep (float, optional): . Defaults to 0.
	deviceId (str, optional): Định danh của thiết bị đặt lệnh
	userAgent (str, optional): Người dùng
	```
	"""
	fc_req = fcmodel_requests.NewOrder(str(account).upper()
	, str(random.randint(0, 99999999))
	, str(instrumentID).upper(), str(market).upper(), str(buySell).upper(), str(orderType).upper()
	, float(price), int(quantity), bool(stopOrder), float(stopPrice), str(stopType), float(stopStep)
 	, float(lossStep), float(profitStep),deviceId= str(deviceId), userAgent = str(userAgent))
	

	res = client.new_order(fc_req)
	return res

@app.get("/modifyOrder")
async def fc_modify_order(orderID: str, instrumentID: str, marketID: str, buySell: str, orderType: str
    , price: float, quantity: int, account: str, deviceId: str = FCTradingClient.get_deviceid(), userAgent: str = FCTradingClient.get_user_agent()):
	"""Modify order

	Args:
	```
	orderID (str): OrderID to modify
	instrumentID (str): Mã chứng khoán
	marketID (str): Thị trường ('VN' for stock 'VNFE' for derviratives)
	buySell (str): 'B' or 'S'
	orderType (str): Loại lệnh
	price (float): Giá
	quantity (int): Khối lượng
	account (str): Tài khoản
	deviceId (str, optional): Định danh thiết bị
	userAgent (str, optional): Định danh người dùng
	```
	Returns:
		Str: json string response
	"""
	fc_rq = fcmodel_requests.ModifyOrder(str(account)
	, str(random.randint(0, 99999999)), str(orderID)
	, str(marketID), str(instrumentID), float(price), int(quantity), str(buySell), str(orderType), deviceId= str(deviceId), userAgent= str(userAgent))

	res = client.modify_order(fc_rq)
	return res

@app.get("/cancelOrder")
async def fc_cancel_order(orderID: str, instrumentID: str, marketID: str, buySell: str, account: str, deviceId: str = FCTradingClient.get_deviceid(), userAgent: str = FCTradingClient.get_user_agent()):
	"""Cancel order

	Args:
	```
	orderID (str): Số hiệu lệnh cần hủy
	instrumentID (str): Mã chứng khoán
	marketID (str): Thị trường 'VN' (cơ sở) hay 'VNFE' (phái sinh)
	buySell (str): 'B' or 'S'
	account (str): Tài khoản
	deviceId (str, optional): Định danh thiết bị đặt lệnh. Defaults to ''.
	userAgent (str, optional): Định danh người dùng. Defaults to ''.
	```
	Returns:
		str: JSON string
	"""
	fc_rq = fcmodel_requests.CancelOrder(str(account), str(random.randint(0, 99999999))
	, str(orderID), str(marketID), str(instrumentID), str(buySell), deviceId=str(deviceId), userAgent=str(userAgent))

	res = client.cancle_order(fc_rq)
	return res

@app.get("/derNewOrder")
async def fc_der_new_order(instrumentID: str, market: str, buySell: str, orderType: str
    , price: float, quantity: int, account: str, stopOrder: bool = False, stopPrice: float = 0, 
      stopType: str = '', stopStep: float = 0, lossStep: float = 0, profitStep: float = 0, deviceId: str = FCTradingClient.get_deviceid(), userAgent: str = FCTradingClient.get_user_agent()):
	
	"""Place new derivatives order

	Args:
	```	
	instrumentID (str): Mã chứng khoán
	market (str): Thị trường ('VN' hoặc 'VNFE')
	buySell (str): 'B' or 'S'
	orderType (str): Loại lệnh
	price (float): Giá. Với các lệnh điều kiện price=0
	quantity (int): Khối lượng
	account (str): Tài khoản
	stopOrder (bool, optional): Lệnh điều kiện (chỉ áp dụng với phái sinh). Defaults to False.
	stopPrice (float, optional): Giá trigger của lệnh điều kiện. Defaults to 0.
	stopType (str, optional): Loại lệnh điều kiện. Defaults to ''.
	stopStep (float, optional): . Defaults to 0.
	lossStep (float, optional): . Defaults to 0.
	profitStep (float, optional): . Defaults to 0.
	deviceId (str, optional): Định danh của thiết bị đặt lệnh
	userAgent (str, optional): Người dùng
	```
	"""
	fc_req = fcmodel_requests.NewOrder(str(account).upper()
	, str(random.randint(0, 99999999))
	, str(instrumentID).upper(), str(market).upper(), str(buySell).upper(), str(orderType).upper()
	, float(price), int(quantity), bool(stopOrder), float(stopPrice), str(stopType), float(stopStep)
 	, float(lossStep), float(profitStep),deviceId= str(deviceId), userAgent = str(userAgent))
	

	res = client.der_new_order(fc_req)
	return res

@app.get("/derModifyOrder")
async def fc_der_modify_order(orderID: str, instrumentID: str, marketID: str, buySell: str, orderType: str
    , price: float, quantity: int, account: str, deviceId: str = FCTradingClient.get_deviceid(), userAgent: str = FCTradingClient.get_user_agent()):
	"""Derivative Modify order

	Args:
	```
	orderID (str): OrderID to modify
	instrumentID (str): Mã chứng khoán
	marketID (str): Thị trường ('VN' for stock 'VNFE' for derviratives)
	buySell (str): 'B' or 'S'
	orderType (str): Loại lệnh
	price (float): Giá
	quantity (int): Khối lượng
	account (str): Tài khoản
	deviceId (str, optional): Định danh thiết bị
	userAgent (str, optional): Định danh người dùng
	```
	Returns:
		Str: json string response
	"""
	fc_rq = fcmodel_requests.ModifyOrder(str(account)
	, str(random.randint(0, 99999999)), str(orderID)
	, str(marketID), str(instrumentID), float(price), int(quantity), str(buySell), str(orderType), deviceId= str(deviceId), userAgent= str(userAgent))

	res = client.der_modify_order(fc_rq)
	return res

@app.get("/derCancelOrder")
async def fc_der_cancel_order(orderID: str, instrumentID: str, marketID: str, buySell: str, account: str, deviceId: str = FCTradingClient.get_deviceid(), userAgent: str = FCTradingClient.get_user_agent()):
	"""Derivative cancel order

	Args:
	```
	orderID (str): Số hiệu lệnh cần hủy
	instrumentID (str): Mã chứng khoán
	marketID (str): Thị trường 'VN' (cơ sở) hay 'VNFE' (phái sinh)
	buySell (str): 'B' or 'S'
	account (str): Tài khoản
	deviceId (str, optional): Định danh thiết bị đặt lệnh. Defaults to ''.
	userAgent (str, optional): Định danh người dùng. Defaults to ''.
	```
	Returns:
		str: JSON string
	"""
	fc_rq = fcmodel_requests.CancelOrder(str(account), str(random.randint(0, 99999999))
	, str(orderID), str(marketID), str(instrumentID), str(buySell), deviceId=str(deviceId), userAgent=str(userAgent))

	res = client.der_cancle_order(fc_rq)
	return res

@app.get("/stockAccountBalance")
async def fc_stock_account_balance(account: str):
	"""Get stock account balance

	Args:
	```
 	account (str): stock account id
  	```
	Returns:
	```
 	str: Json str
  	```
	"""	
	
	fc_rq = fcmodel_requests.StockAccountBalance(str(account))

	res = client.get_stock_account_balance(fc_rq)
	return res

@app.get("/derivativeAccountBalance")
async def fc_derivative_account_balance(account: str):
	"""Get derivative account balance

	Args:
		account (str): derivative account id

	Returns:
		str: Json str
	"""	
	
	fc_rq = fcmodel_requests.DerivativeAccountBalance(str(account))

	res = client.get_derivative_account_balance(fc_rq)
	return res

@app.get("/ppmmrAccount")
async def fc_pp_mmr_account(account: str):
	"""Get pp and mmr of account

	Args:
		account (str): stock account id

	Returns:
		str: Json str
	"""	
	
	fc_rq = fcmodel_requests.PPMMRAccount(str(account))

	res = client.get_pp_mmr_account(fc_rq)
	return res

@app.get("/stockPosition")
async def fc_stock_position(account: str):
	"""Get stock position of account

	Args:
		account (str): stock account id

	Returns:
		str: Json str
	"""	
	
	fc_rq = fcmodel_requests.StockPosition(str(account))

	res = client.get_stock_position(fc_rq)
	return res

@app.get("/derivativePosition")
async def fc_derivative_position(account: str):
	"""Get derivative position of account

	Args:
		account (str): derivative account id

	Returns:
		str: Json str
	"""	
	
	fc_rq = fcmodel_requests.DerivativePosition(str(account))

	res = client.get_derivative_position(fc_rq)
	return res

@app.get("/maxBuyQty")
async def fc_max_buy_qty(account: str, instrumentID: str, price: float):
	"""Get max buy qty

	Args:
		account (str): account id
		instrumentID (str): buy instrument
		price (float): buy price

	Returns:
		str: json string
	"""
	
	fc_rq = fcmodel_requests.MaxBuyQty(str(account), str(instrumentID), float(price))

	res = client.get_max_buy_qty(fc_rq)
	return res

@app.get("/maxSellQty")
async def fc_max_sell_qty(account: str, instrumentID: str, price: float =0):
	"""Get max sell qty

	Args:
		account (str): account id
		instrumentID (str): sell instrument
		price (float): sell price ( for dervatives exchange )

	Returns:
		str: json string
	"""
	
	fc_rq = fcmodel_requests.MaxSellQty(str(account), str(instrumentID), float(price))

	res = client.get_max_sell_qty(fc_rq)
	return res

@app.get("/orderHistory")
async def fc_order_history(account: str, startDate: str, endDate: str):
	"""Get order history

	Args:
		account (str): account id
		startDate (str): from date (format 'DD/MM/YYYY' ex: 21/05/2021)
		endDate (str): end date (format 'DD/MM/YYYY' ex: 21/05/2021)

	Returns:
		[str]: json string
	"""
	
	fc_rq = fcmodel_requests.OrderHistory(str(account), str(startDate), str(endDate))

	res = client.get_order_history(fc_rq)
	return res

@app.get("/orderBook")
async def fc_order_book(account: str):
	"""Get order book

	Args:
		account (str): account id

	Returns:
		[str]: json string
	"""
	
	fc_rq = fcmodel_requests.OrderBook(str(account))

	res = client.get_order_book(fc_rq)
	return res

@app.get("/auditOrderBook")
async def fc_audit_order_book(account: str):
	"""Get audit order book ( include order error)

	Args:
		account (str): account id

	Returns:
		[str]: json string
	"""
	
	fc_rq = fcmodel_requests.AuditOrderBook(str(account))

	res = client.get_audit_order_book(fc_rq)
	return res

@app.get("/rateLimit")
async def fc_rateLimit():
	"""Get ratelimit

	Returns:
		[str]: json string
	"""
	
	res = client.get_ratelimit()
	return res

@app.get("/cash/cashInAdvanceAmount", tags=["cash"])
async def fc_cia_amount(models: fcmodel_requests.CashInAdvanceAmount = Depends()):

	res = client.get_cash_cia_amount(models)
	return res

@app.get("/cash/unsettleSoldTransaction", tags=["cash"])
async def fc_cash_unsettleSoldTransaction(models: fcmodel_requests.UnsettleSoldTransaction = Depends()):

	res = client.get_cash_unsettle_sold_transaction(models)
	return res
@app.get("/cash/transferHistories", tags=["cash"])
async def fc_cash_transferHistories(models: fcmodel_requests.CashTransferHistory = Depends()):

	res = client.get_cash_cia_amount(models)
	return res
@app.get("/cash/cashInAdvanceHistories", tags=["cash"])
async def fc_cash_cashInAdvanceHistories(models: fcmodel_requests.CashInAdvanceHistory = Depends()):

	res = client.get_cash_cia_history(models)
	return res
@app.get("/cash/estCashInAdvanceFee", tags=["cash"])
async def fc_cash_estCashInAdvanceFee(models: fcmodel_requests.CashInAdvanceEstFee = Depends()):

	res = client.get_cash_cia_est_fee(models)
	return res
@app.post("/cash/vsdCashDW", tags=["cash"])
async def fc_cash_vsdCashDW(models: fcmodel_requests.CashTransferVSD):

	res = client.create_cash_transfer_vsd(models)
	return res
@app.post("/cash/transferInternal", tags=["cash"])
async def fc_cash_transferInternal(models: fcmodel_requests.CashTransfer):

	res = client.create_cash_transfer(models)
	return res
@app.post("/cash/createCashInAdvance", tags=["cash"])
async def fc_cash_createCashInAdvance(models: fcmodel_requests.CashCIA):

	res = client.create_cia(models)
	return res

#ORS
@app.get("/ors/dividend", tags=["online right subscription"])
async def fc_ors_dividend(models: fcmodel_requests.OrsDividend = Depends()):

	res = client.get_ors_dividend(models)
	return res
@app.get("/ors/exercisableQuantity", tags=["online right subscription"])
async def fc_ors_exercisableQuantity(models: fcmodel_requests.OrsExercisableQuantity = Depends()):

	res = client.get_ors_exercisable_quantity(models)
	return res
@app.get("/ors/histories", tags=["online right subscription"])
async def fc_ors_histories(models: fcmodel_requests.OrsHistory = Depends()):

	res = client.get_ors_history(models)
	return res
@app.post("/ors/create", tags=["online right subscription"])
async def fc_ors_create(models: fcmodel_requests.Ors):

	res = client.create_ors(models)
	return res

#STOCK
@app.get("/stock/transferable", tags=["stock"])
async def fc_stock_exercisableQuantity(models: fcmodel_requests.StockTransferable = Depends()):

	res = client.get_stock_transferable(models)
	return res
@app.get("/stock/transferHistories", tags=["stock"])
async def fc_stock_histories(models: fcmodel_requests.StockTransferHistory = Depends()):

	res = client.get_stock_transfer_history(models)
	return res
@app.post("/stock/transfer", tags=["stock"])
async def fc_stock_create(models: fcmodel_requests.StockTransfer):

	res = client.create_stock_transfer(models)
	return res



from ..events import EventHook


class Hub:
    def __init__(self, name, connection):
        self.name = name
        self.server = HubServer(name, connection, self)
        self.client = HubClient(name, connection)
        self.error = EventHook()


class HubServer:
    def __init__(self, name, connection, hub):
        self.name = name
        self.__connection = connection
        self.__hub = hub

    def invoke(self, method, *data):
        self.__connection.send({
            'H': self.name,
            'M': method,
            'A': data,
            'I': self.__connection.increment_send_counter()
        })


class HubClient(object):
    def __init__(self, name, connection):
        self.name = name
        self.__handlers = {}

        def handle(**kwargs):
            messages = kwargs['M'] if 'M' in kwargs and len(kwargs['M']) > 0 else {}
            for inner_data in messages:
                hub = inner_data['H'] if 'H' in inner_data else ''
                if hub.lower() == self.name.lower():
                    method = inner_data['M']
                    if method in self.__handlers:
                        arguments = inner_data['A']
                        self.__handlers[method].fire(*arguments)

        connection.received += handle

    def on(self, method, handler):
        if method not in self.__handlers:
            self.__handlers[method] = EventHook()
        self.__handlers[method] += handler

    def off(self, method, handler):
        if method in self.__handlers:
            self.__handlers[method] -= handler


class DictToObj:
    def __init__(self, d):
        self.__dict__ = d



import configparser
import os
# READ CONFIG
config = configparser.SafeConfigParser(os.environ)
config.read(os.path.dirname(os.path.abspath(__file__)) + "/config.ini")

ConsumerID     = config.get("DEFAULT", "ConsumerID", fallback="")
ConsumerSecret = config.get("DEFAULT", "ConsumerSecret", fallback="")
PrivateKey     = config.get("DEFAULT", "PrivateKey", fallback="")
Url            = config.get("DEFAULT", "Url", fallback="https://fc-tradeapi.ssi.com.vn/") 
StreamURL      = config.get("DEFAULT", "StreamURL", fallback="https://fc-tradehub.ssi.com.vn/")
TwoFAType      = int(config.get("DEFAULT", "TwoFAType", fallback="0")) # 0-PIN, 1-OTP
NotifyId       = config.get("DEFAULT", "NotifyId", fallback="-1")



SIGNALR = "v2.0/signalr"
SIGNALR_METHOD = "Broadcast"
SIGNALR_METHOD_ERROR = "Error"
SIGNALR_HUB_FC = "BroadcastHubV2"

FC_GET_OTP                          = "api/v2/Trading/GetOTP"
FC_GET_ACCESS_TOKEN                 = "api/v2/Trading/AccessToken"

FC_NEW_ORDER                        = "api/v2/Trading/NewOrder"
FC_MODIFY_ORDER                     = "api/v2/Trading/ModifyOrder"
FC_CANCEL_ORDER                     = "api/v2/Trading/CancelOrder"

FC_DER_NEW_ORDER                    = "api/v2/Trading/derNewOrder"
FC_DER_MODIFY_ORDER                 = "api/v2/Trading/derModifyOrder"
FC_DER_CANCEL_ORDER                 = "api/v2/Trading/derCancelOrder"

FC_GET_ORDER_HISTORY                = "api/v2/Trading/orderHistory"
FC_GET_ORDER_BOOK                   = "api/v2/Trading/orderBook"
FC_GET_AUDIT_ORDER_BOOK             = "api/v2/Trading/auditOrderBook"

FC_GET_DER_POSITION                 = "api/v2/Trading/derivPosition"
FC_GET_STOCK_POSITION               = "api/v2/Trading/stockPosition"
FC_GET_MAX_BUY_QUANTITY             = "api/v2/Trading/maxBuyQty"
FC_GET_MAX_SELL_QUANTITY            = "api/v2/Trading/maxSellQty"

FC_GET_ACCOUNT_BALANCE              = "api/v2/Trading/cashAcctBal"
FC_GET_DER_ACCOUNT_BALANCE          = "api/v2/Trading/derivAcctBal"
FC_GET_PP_MMR_ACCOUNT               = "api/v2/Trading/ppmmraccount"

FC_GET_RATELIMIT                    = "api/v2/Trading/rateLimit"

#CASH API
FC_CASH_CIA_AMOUNT                  = "api/v2/cash/cashInAdvanceAmount"
FC_CASH_UNSETTLE_SOLD_TRANSACTION   = "api/v2/cash/unsettleSoldTransaction"
FC_CASH_TRANSFER_HISTORY            = "api/v2/cash/transferHistories"
FC_CASH_CIA_HISTORY                 = "api/v2/cash/cashInAdvanceHistories"
FC_CASH_CIA_EST_FEE                 = "api/v2/cash/estCashInAdvanceFee"
FC_CASH_VSD_DW                      = "api/v2/cash/vsdCashDW"
FC_CASH_TRANSFER                    = "api/v2/cash/transferInternal"
FC_CASH_CIA                         = "api/v2/cash/createCashInAdvance"

#ONLINE RIGHT SUBSCRIPTION
FC_ORS_DIVIDEND                     = "api/v2/ors/dividend"
FC_ORS_EXCERCISABLE_QTY             = "api/v2/ors/exercisableQuantity"
FC_ORS_HISTORY                      = "api/v2/ors/histories"
FC_ORS                              = "api/v2/ors/create"

#STOCK
FC_STOCK_TRANSFERABLE               = "api/v2/stock/transferable"
FC_STOCK_HISTORY                    = "api/v2/stock/transferHistories"
FC_STOCK_TRANSFER                   = "api/v2/stock/transfer"




import websocket
import threading
import requests
import traceback
import time
import ssl
from .reconnection import ConnectionStateChecker
from .connection import ConnectionState
from ..hubs.errors import HubError, HubConnectionError, UnAuthorizedHubError
from .base_transport import BaseTransport
from ..helpers import Helpers
from json import dumps, loads
from urllib.parse import urlparse, urlunparse, urlencode, quote

class WebsocketTransport(BaseTransport):
    def __init__(self,
            url="",
            headers=None,
            keep_alive_interval=15,
            reconnection_handler=None,
            verify_ssl=False,
            skip_negotiation=False,
            enable_trace=False,
            **kwargs):
        super(WebsocketTransport, self).__init__(**kwargs)
        self.protocol_version = '1.3'
        self._ws = None
        self.enable_trace = enable_trace
        self._thread = None
        self.skip_negotiation = skip_negotiation
        self.url = url
        if headers is None:
            self.headers = dict()
        else:
            self.headers = headers
        self.handshake_received = False
        self.token = None  # auth
        self.state = ConnectionState.disconnected
        self.connection_alive = False
        self._thread = None
        self._ws = None
        self.verify_ssl = verify_ssl
        self.conn_data = ''
        self.connection_checker = ConnectionStateChecker(
            lambda: self.logger.debug("Ping"),
            keep_alive_interval
        )
        self.reconnection_handler = reconnection_handler

        if len(self.logger.handlers) > 0:
            websocket.enableTrace(self.enable_trace, self.logger.handlers[0])
    
    def is_running(self):
        return self.state != ConnectionState.disconnected

    def stop(self):
        if self.state == ConnectionState.connected:
            self.connection_checker.stop()
            self._ws.close()
            self.state = ConnectionState.disconnected
            self.handshake_received = False

    def start(self, connectionData = ''):
        self.conn_data = connectionData
        if not self.skip_negotiation:
            self._negotiate()

        if self.state == ConnectionState.connected:
            self.logger.warning("Already connected unable to start")
            return False

        self.state = ConnectionState.connecting
        self.logger.debug("start url:" + self.url)
        self._ws = websocket.WebSocketApp(
            self.url,
            header=self.headers,
            on_message=self.on_message,
            on_error=self.on_socket_error,
            on_close=self.on_close,
            on_open=self.on_open,
            )
            
        self._thread = threading.Thread(
            target=lambda: self._ws.run_forever(
                sslopt={"cert_reqs": ssl.CERT_NONE}
                if not self.verify_ssl else {}
            ))
        self._thread.daemon = True
        self._thread.start()
        return True

    @staticmethod
    def _format_url(url, action, query):
        return '{url}/{action}?{query}'.format(url=url, action=action, query=query)
    
    def _get_ws_url_from(self):
        parsed = urlparse(self.url)
        scheme = 'wss' if parsed.scheme == 'https' else 'ws'
        url_data = (scheme, parsed.netloc, parsed.path, parsed.params, parsed.query, parsed.fragment)

        return urlunparse(url_data)
    @staticmethod
    def _get_cookie_str(request):
        return '; '.join([
            '%s=%s' % (name, value)
            for name, value in request.items()
        ])
    def _get_socket_url(self, responseData):
        ws_url = self._get_ws_url_from()
        query = urlencode({
            'transport': 'webSockets',
            'connectionToken':responseData['ConnectionToken'],
            'connectionData': self.conn_data,
            'clientProtocol': responseData['ProtocolVersion'],
        }, quote_via=quote)

        return self._format_url(ws_url, 'connect', query)
    def _negotiate(self):
        query = urlencode({
            'connectionData': self.conn_data,
            'clientProtocol': self.protocol_version,
        }, quote_via=quote)
        negotiate_url = self._format_url(self.url, 'negotiate', query)
        self.logger.debug("Negotiate url:{0}".format(negotiate_url))

        response = requests.post(
            negotiate_url, headers=self.headers, verify=self.verify_ssl)
        self.logger.debug(
            "Response status code{0}".format(response.status_code))
        
        if response.status_code != 200:
            raise HubError(response.status_code)\
                if response.status_code != 401 else UnAuthorizedHubError()

        data = response.json()
        # self.headers['Cookie'] = self._get_cookie_str(response.cookies)
        self.url = self._get_socket_url(data)

       

    def on_open(self, _):
        self.logger.debug("-- web socket open --")
        self._on_open()

    def on_close(self, callback, close_status_code, close_reason):
        self.logger.debug("-- web socket close --")
        self.logger.debug(close_status_code)
        self.logger.debug(close_reason)
        self.state = ConnectionState.disconnected
        if self._on_close is not None and callable(self._on_close):
            self._on_close()
        if callback is not None and callable(callback):
            callback()

    def on_reconnect(self):
        self.logger.debug("-- web socket reconnecting --")
        self.state = ConnectionState.disconnected
        if self._on_close is not None and callable(self._on_close):
            self._on_close()

    def on_socket_error(self, app, error):
        """
        Args:
            _: Required to support websocket-client version equal or greater than 0.58.0
            error ([type]): [description]

        Raises:
            HubError: [description]
        """
        self.logger.debug("-- web socket error --")
        self.logger.error(traceback.format_exc(10, True))
        self.logger.error("{0} {1}".format(self, error))
        self.logger.error("{0} {1}".format(error, type(error)))
        self._on_close()
        self.state = ConnectionState.disconnected
        #raise HubError(error)
        

    def on_message(self, app, raw_message):
        self.logger.debug("Message received{0}".format(raw_message))
        if len(raw_message) > 0:
            return self._on_message(raw_message)

    def send(self, message):
        self.logger.debug("Sending message {0}".format(message))
        try:
            self._ws.send(dumps(message))
            self.connection_checker.last_message = time.time()
            if self.reconnection_handler is not None:
                self.reconnection_handler.reset()
        except (
                websocket._exceptions.WebSocketConnectionClosedException,
                OSError) as ex:
            self.handshake_received = False
            self.logger.warning("Connection closed {0}".format(ex))
            self.state = ConnectionState.disconnected
            if self.reconnection_handler is None:
                if self._on_close is not None and\
                        callable(self._on_close):
                    self._on_close()
                raise ValueError(str(ex))
            # Connection closed
            self.handle_reconnect()
        except Exception as ex:
            raise ex

    def handle_reconnect(self):
        if not self.reconnection_handler.reconnecting and self._on_reconnect is not None and \
                callable(self._on_reconnect):
            self._on_reconnect()
        self.reconnection_handler.reconnecting = True
        try:
            self.stop()
            self.start()
        except Exception as ex:
            self.logger.error(ex)
            sleep_time = self.reconnection_handler.next()
            threading.Thread(
                target=self.deferred_reconnect,
                args=(sleep_time,)
            ).start()

    def deferred_reconnect(self, sleep_time):
        time.sleep(sleep_time)
        


from ssi_fctrading import FCTradingStream, fc_stream
from ssi_fctrading import FCTradingClient
from urllib.parse import urlparse, urlunparse, urlencode, quote

import urllib3
from . import fc_config


def on_message(tapi_message):
    print("fc_received: " + tapi_message)


def on_error(tapi_error):
    print("fc_error: " + tapi_error)

def on_open():
    print('Connected to ' + fc_config.StreamURL)

def tapi_data_streaming(on_message, on_error):
    client = FCTradingClient(fc_config.Url, fc_config.ConsumerID,
                             fc_config.ConsumerSecret, fc_config.PrivateKey, fc_config.TwoFAType)
    print("access_token: " + client.get_access_token())
    stream_client = FCTradingStream(client, fc_config.StreamURL, on_message, on_error, fc_config.NotifyId, on_open=on_open)
    stream_client.start()
    message = None
    while message != "exit()":
        message = input(">> ")


# main function
if __name__ == '__main__':
    tapi_data_streaming(on_message, on_error)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# signalr_aio/_connection.py
# Stanislav Lazarov


from .events import EventHook
from .hubs import Hub
from .transports import WebsocketTransport
from .helpers import Helpers
from json import dumps, loads
import logging

class Connection(object):
  protocol_version = '1.3'

  def __init__(self, url, headers=None,
            **kwargs):
    self.url = url
    self.__hubs = {}
    self.__send_counter = -1
    self.hub = None
    self.session = None
    self.headers = headers
    self.received = EventHook()
    self.error = EventHook()
    self.started = False
    self.logger = Helpers.get_logger()
    self.__transport = WebsocketTransport(self.url, self.headers, on_message=self._on_message, **kwargs )
    
    def handle_error(**kwargs):
      error = kwargs["E"] if "E" in kwargs else None
      if error is None:
          return

      self.error.fire(error)

    self.received += handle_error

  def start(self):
      cdata = self._get_conn_data(self.__hubs)
      self.logger.debug("Connection started")
      return self.__transport.start(connectionData=cdata)
  @staticmethod
  def _get_conn_data(hubs):
      conn_data = dumps([{'name': hub for hub in hubs}])
      return conn_data
  
  def on_open_callback(self, callback):
      self.__transport.on_open_callback(callback=callback)
  
  def _on_message(self, data):
      msg = loads(data)
      self.received.fire(**msg)
  
  def on_open(self, data):
      msg = loads(data)
      self.received.fire(**msg)
      
  def register_hub(self, name):
      if name not in self.__hubs:
          if self.started:
              raise RuntimeError(
                  'Cannot create new hub because connection is already started.')
          self.__hubs[name] = Hub(name, self)
          return self.__hubs[name]

  def increment_send_counter(self):
      self.__send_counter += 1
      return self.__send_counter

  def send(self, message):
      self.__transport.send(message)

  def close(self):
      self.__transport.close()



from ..helpers import Helpers

class BaseTransport(object):
    def __init__(self, on_message=None, on_open=None, on_close=None):
        self._on_message= on_message
        self.logger = Helpers.get_logger()
        self._on_open = on_open if on_open else lambda: self.logger.info("on_connect not defined")
        self._on_close = on_close if on_close else lambda: self.logger.info("on_disconnect not defined")
        self._on_reconnect = lambda: self.logger.info("on_reconnect not defined")

    def on_open_callback(self, callback):
        self._on_open = callback
    
    def on_close_callback(self, callback):
        self._on_close = callback

    def on_reconnect_callback(self, callback):
        self._on_reconnect = callback

    def start(self, connectionData = ''): # pragma: no cover
        raise NotImplementedError()

    def stop(self): # pragma: no cover
        raise NotImplementedError()
    
    def is_running(self): # pragma: no cover
        raise NotImplementedError()

    def send(self, message, on_invocation = None): # pragma: no cover
        raise NotImplementedError()